package org.processmining.stream.utils;

import java.awt.Point;
import java.awt.Rectangle;
import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;
import org.processmining.models.heuristics.HeuristicsNet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.heuristicsnet.miner.heuristics.converter.HeuristicsNetToPetriNetConverter;
import org.processmining.plugins.pnml.exporting.PnmlExportNetToPNML;

/**
 * A collection of user interface utilities
 * 
 * @author Andrea Burattin
 */
public class Utils {

	/**
	 * This method tells whether the given event is the tagged as the first
	 * event of a stream
	 * 
	 * @param e an event
	 * @return <tt>true</tt> if the event is tagged as the first event of a
	 * trace, <tt>false</tt> otherwise
	 */
	public static boolean isTraceStart(XEvent e) {
		if (e.getAttributes().containsKey("stream:lifecycle:trace-transition")) {
			return ((XAttributeLiteral) e.getAttributes().get("stream:lifecycle:trace-transition")).getValue().equals("start");
		}
		return false;
	}
	
	/**
	 * This method tells whether the given event is the tagged as the last
	 * event of a stream
	 * 
	 * @param e an event
	 * @return <tt>true</tt> if the event is tagged as the last event of a
	 * trace, <tt>false</tt> otherwise
	 */
	public static boolean isTraceComplete(XEvent e) {
		if (e.getAttributes().containsKey("stream:lifecycle:trace-transition")) {
			return ((XAttributeLiteral) e.getAttributes().get("stream:lifecycle:trace-transition")).getValue().equals("complete");
		}
		return false;
	}
	
	/**
	 * This method converts a time duration, expressed in milliseconds, into a
	 * human readable version (e.g. "5 seconds", "6,7 hours", "12 days").
	 * 
	 * @param duration the time duration, expressed in milliseconds
	 * @return a string representation of the duration
	 */
	public static String fromMillisecToStringDuration(long duration) {
		DecimalFormat twoDForm = new DecimalFormat("#.##");
		if (duration < 0) {
			return "Error";
		} else if (duration == 0) {
			return "0 milliseconds";
		} else if (duration < 100) {
			return duration + " milliseconds";
		} else if (duration < 60000) {			
			return twoDForm.format(duration/1000.0) + " seconds";
		} else if (duration < 3600000) {
			return twoDForm.format(duration/60000.0) + " minutes";
		} else if (duration < 216000000) {
			return twoDForm.format(duration/3600000.0) + " hours";
		} else {
			double hours = duration/3600000.0; // really necessary? we are getting out of int
			return twoDForm.format(hours/24.0) + " days";
		}
	}
	
	/**
	 * This method check if, given a value, it's inside the interval defined by
	 * its left and right area
	 * 
	 * @param value the value to be checked
	 * @param left the leftmost bound
	 * @param right the rightmost bound
	 * @return true if the given value is inside the interval
	 */
	public static boolean inRange(int value, int left, int right) {
		return value >= left && value <= right;
	}
	
	/**
	 * This method check if, given a point, it's inside the given rectangle
	 * 
	 * @param p the point to be checked
	 * @param r the rectangle bound
	 * @return true if the given value is inside the interval
	 */
	public static boolean inRectangle(Point p, Rectangle r) {
		return inRange(p.x, r.x, r.x + r.width) && inRange(p.y, r.y, r.y + r.height);
	}
	
	/**
	 * This method, given a trace, returns its case id
	 * 
	 * @param t the given trace
	 * @return the case id associated with the trace
	 */
	public static String getCaseID(XTrace t) {
		return t.getAttributes().get("concept:name").toString();
	}
	
	/**
	 * This method, given an event, returns the name of the activity
	 * 
	 * @param e the log event
	 * @return the name of the activity of the event
	 */
	public static String getActivityName(XEvent e) {
		return e.getAttributes().get("concept:name").toString();
	}
	
	/**
	 * 
	 * @param context
	 * @param net
	 * @return
	 */
	public static Petrinet toPetri(PluginContext context, HeuristicsNet net) {
		return (Petrinet) HeuristicsNetToPetriNetConverter.converter(context, net)[0];
	}
	
	/**
	 * 
	 * @param net
	 * @param startName
	 * @param endName
	 * @return
	 */
	public static Object[] addArtificialStartEnd(PluginContext context, Petrinet net, String startName, String endName) {
		HashMap<String, Transition> transitionMap = new HashMap<String, Transition>();
		HashMap<String, Place> placeMap = new HashMap<String, Place>();
		
		Petrinet newNet = new PetrinetImpl(net.getLabel());
		Marking m = new Marking();
		
		// create artificial start transition and places
		Place startPlace = newNet.addPlace("startplace");
		Transition start = newNet.addTransition(startName);
		Place afterStart = newNet.addPlace("afterstrat");
		newNet.addArc(startPlace, start);
		newNet.addArc(start, afterStart);
		
		m.add(startPlace);
		
		// add all transitions, places and edges
		for (Transition t : net.getTransitions()) {
			transitionMap.put(t.getId().toString(), newNet.addTransition(t.getLabel()));
		}
		for (Place p : net.getPlaces()) {
			placeMap.put(p.getId().toString(), newNet.addPlace(p.getId().toString()));
		}
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : net.getEdges()) {
			if (e.getSource() instanceof Transition) {
				newNet.addArc(transitionMap.get(((Transition)e.getSource()).getId().toString()), placeMap.get(((Place)e.getTarget()).getId().toString()));
			} else {
				newNet.addArc(placeMap.get(((Place)e.getSource()).getId().toString()), transitionMap.get(((Transition)e.getTarget()).getId().toString()));
			}
		}
		
		// create artificial end transition and places
		Place beforeEndPlace = newNet.addPlace("beforeend");
		Transition end = newNet.addTransition(endName);
		Place endPlace = newNet.addPlace("endplace");
		newNet.addArc(beforeEndPlace, end);
		newNet.addArc(end, endPlace);
		
		for (Place p : newNet.getPlaces()) {
			if (!(p.getLabel().equals("startplace") || p.getLabel().equals("endplace") ||
					p.getLabel().equals("afterstrat")) || p.getLabel().equals("beforeend")) {
				if (newNet.getInEdges(p).isEmpty()) {
					Transition localStart = newNet.addTransition("");
					localStart.setInvisible(true);
					newNet.addArc(afterStart, localStart);
					newNet.addArc(localStart, p);
				}
				if (newNet.getOutEdges(p).isEmpty()) {
					Transition localEnd = newNet.addTransition("");
					localEnd.setInvisible(true);
					newNet.addArc(p, localEnd);
					newNet.addArc(localEnd, beforeEndPlace);
				}
			}
		}
		
		context.addConnection(new InitialMarkingConnection(newNet, m));
		return new Object[] { newNet, m };
	}
	
	/**
	 * Returns the harmonic mean of the given list of values
	 * 
	 * @see org.processmining.plugins.joosbuijs.blockminer.genetic.TreeEvaluator
	 * @author jbuijs
	 * @param values the list of values
	 * @return the harmonic mean
	 */
	public static double harmonicMean(List<Double> values) {
		//A list of 1 returns 1.0
		if (values.size() == 1) {
			return values.get(0);
		}
		double sum = 0;
		double product = 1;
		for (Double dub : values) {
			sum += dub;
			product *= dub;
		}

		double numerator = values.size() * product;
		double denominator = sum;
		if (denominator == 0)
			return 0;
		return numerator / denominator;
	}
	
	/**
	 * This method parses a string (typically given as argument on the command
	 * line) like "(a,b);(c,d);(e,f)    ; ( ggg  ,  hhh)" and generates an array
	 * list of pairs of string
	 * 
	 * @param argument
	 * @return
	 */
	public static ArrayList<Pair<String, String>> parseString(String argument) {
		ArrayList<Pair<String, String>> toReturn = new ArrayList<Pair<String, String>>();
		
		Pattern p1 = Pattern.compile("\\s*;\\s*");
		Pattern p2 = Pattern.compile("\\(\\s*(\\w*)\\s*,\\s*(\\w*)\\s*\\)");
		
		for(String s : p1.split(argument)) {
			Matcher m = p2.matcher(s);
			if (m.matches()) {
				toReturn.add(new Pair<String, String>(m.group(1), m.group(2)));
			}
		}
		
		return toReturn;
	}
	
	/**
	 * 
	 * @param context
	 * @param destinationPath
	 * @param modelName
	 * @param model
	 */
	public static void saveToFile(UIPluginContext context, String destinationPath, String modelName, HeuristicsNet model) {
		if (destinationPath == null || modelName == null) {
			return;
		}
		String completePath = destinationPath  + System.getProperty("file.separator") + modelName;
		try {
			Petrinet p = Utils.toPetri(context, model);
			PnmlExportNetToPNML xport = new PnmlExportNetToPNML();
			xport.exportPetriNetToPNMLFile(context, p, new File(completePath));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
