package org.processmining.streamheuristicsminer.plugins;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.util.HashMap;

import javax.swing.DefaultListModel;
import javax.swing.JComponent;
import javax.swing.ListSelectionModel;

import org.deckfour.uitopia.api.event.TaskListener;
import org.deckfour.xes.classification.XEventAndClassifier;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventLifeTransClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.basicutils.parameters.impl.PluginParametersImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.Pair;
import org.processmining.framework.util.ui.widgets.ProMList;
import org.processmining.models.cnet.CNet;
import org.processmining.stream.algorithms.lossycounting.CNetGenerator;
import org.processmining.widgets.wizard.AbstractDialog;
import org.processmining.widgets.wizard.Dialog;
import org.processmining.widgets.wizard.Wizard;
import org.processmining.widgets.wizard.WizardResult;

import com.fluxicon.slickerbox.components.NiceDoubleSlider;
import com.fluxicon.slickerbox.components.NiceIntegerSlider;
import com.fluxicon.slickerbox.components.NiceSlider.Orientation;
import com.fluxicon.slickerbox.factory.SlickerFactory;

@Plugin(name = "Discover C-Net (Stream Heuristics Miner / Baseline)", parameterLabels = {
		"Event Log" }, returnLabels = { "C-Net" }, returnTypes = { CNet.class })
public class DiscoverBaseLineCNetPlugin {

	public class DiscoverBaseLineCNetPluginParametersImpl extends PluginParametersImpl {
		private XEventClassifier classifier = new XEventNameClassifier();
		private double dependencyThreshold = 0.9;
		private double positiveObservations = 10;
		private double andThreshold = 0.05;

		public double getDependencyThreshold() {
			return dependencyThreshold;
		}

		public XEventClassifier getEventClassifier() {
			return classifier;
		}

		public double getPositiveObservations() {
			return positiveObservations;
		}

		public double getAndThreshold() {
			return andThreshold;
		}

		public void setDependencyThreshold(double dependencyThreshold) {
			this.dependencyThreshold = dependencyThreshold;
		}

		public void setEventClassifier(XEventClassifier classifier) {
			this.classifier = classifier;
		}

		public void setPositiveObservations(double positiveObservations) {
			this.positiveObservations = positiveObservations;
		}

		public void setAndThreshold(double andThreshold) {
			this.andThreshold = andThreshold;
		}
	}

	public class DiscoverBaseLineDialogImpl extends AbstractDialog<DiscoverBaseLineCNetPluginParametersImpl> {

		private static final long serialVersionUID = 6489821333751100651L;
		private ProMList<XEventClassifier> classifiersProMList = null;
		private NiceDoubleSlider depThr;
		private NiceDoubleSlider relBest;
		private NiceIntegerSlider posObs;

		private final XLog log;

		public DiscoverBaseLineDialogImpl(UIPluginContext context, String title,
				DiscoverBaseLineCNetPluginParametersImpl parameters,
				Dialog<DiscoverBaseLineCNetPluginParametersImpl> previous, final XLog log) {
			super(context, title, parameters, previous);
			this.log = log;
		}

		protected boolean canProceedToNext() {
			return true;
		}

		private ProMList<XEventClassifier> constructClassifierList(final XLog log) {
			if (classifiersProMList == null) {
				DefaultListModel<XEventClassifier> classifiersList = new DefaultListModel<XEventClassifier>();
				for (XEventClassifier ec : log.getClassifiers()) {
					classifiersList.addElement(ec);
				}
				final XEventClassifier defaultClassifier = new XEventAndClassifier(new XEventNameClassifier(),
						new XEventLifeTransClassifier());
				if (classifiersList.isEmpty()) {
					classifiersList.addElement(defaultClassifier);
					getParameters().setEventClassifier(defaultClassifier);
				}
				classifiersProMList = new ProMList<XEventClassifier>("Select the desired event classifier:",
						classifiersList);
				classifiersProMList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
				classifiersProMList.setPreferredSize(new Dimension(100, 150));
				if (getParameters().getEventClassifier() != null) {
					classifiersProMList.setSelection(getParameters().getEventClassifier());
				} else {
					classifiersProMList.setSelectedIndex(0);
				}
			}
			return classifiersProMList;
		}

		protected Dialog<DiscoverBaseLineCNetPluginParametersImpl> determineNextDialog() {
			return null;
		}

		protected String getUserInputProblems() {
			return "";
		}

		public boolean hasNextDialog() {
			return false;
		}

		public void updateParametersOnGetNext() {
			getParameters().setEventClassifier(classifiersProMList.getSelectedValuesList().get(0));
			getParameters().setDependencyThreshold(depThr.getValue());
			getParameters().setAndThreshold(relBest.getValue());
			getParameters().setPositiveObservations(new Double(posObs.getValue()));
		}

		public void updateParametersOnGetPrevious() {
			updateParametersOnGetNext();
		}

		public JComponent visualize() {
			removeAll();
			constructClassifierList(log);
			// set some layout
			this.setOpaque(false);
			this.setLayout(new GridBagLayout());
			GridBagConstraints c = new GridBagConstraints();
			c.weightx = 1;
			c.fill = GridBagConstraints.HORIZONTAL;
			c.anchor = GridBagConstraints.FIRST_LINE_START;

			// row 0, col 0
			c.gridx = c.gridy = 0;
			add(classifiersProMList, c);

			//		c.gridy += 1;
			//		add(engineList, c);

			c.gridy += 1;
			depThr = SlickerFactory.instance().createNiceDoubleSlider("Dependency Threshold", 0, 1,
					getParameters().getDependencyThreshold(), Orientation.HORIZONTAL);
			add(depThr, c);

			c.gridy += 1;
			relBest = SlickerFactory.instance().createNiceDoubleSlider("And Threshold", 0, 1,
					getParameters().getAndThreshold(), Orientation.HORIZONTAL);
			add(relBest, c);

			c.gridy += 1;
			posObs = SlickerFactory.instance().createNiceIntegerSlider("Positive Observations", 1, 500,
					new Double(getParameters().getPositiveObservations()).intValue(), Orientation.HORIZONTAL);
			add(posObs, c);

			revalidate();
			repaint();
			return this;
		}

	}

	public CNet apply(PluginContext context, XLog eventLog, DiscoverBaseLineCNetPluginParametersImpl parameters) {
		return apply(eventLog, parameters);
	}

	@UITopiaVariant(affiliation = "Eindhoven University of Technology", author = "S.J. van Zelst", email = "s.j.v.zelst@tue.nl")
	@PluginVariant(requiredParameterLabels = { 0 })
	public CNet apply(UIPluginContext context, XLog eventLog) {
		Dialog<DiscoverBaseLineCNetPluginParametersImpl> dialog = new DiscoverBaseLineDialogImpl(context,
				"Heuristics Miner Base Line", new DiscoverBaseLineCNetPluginParametersImpl(), null, eventLog);
		WizardResult<DiscoverBaseLineCNetPluginParametersImpl> wizardResult = Wizard.show(context, dialog);
		if (wizardResult.getInteractionResult().equals(TaskListener.InteractionResult.FINISHED)) {
			DiscoverBaseLineCNetPluginParametersImpl params = wizardResult.getParameters();
			return apply(eventLog, params);
		} else {
			context.getFutureResult(0).cancel(true);
		}
		return null;
	}

	public CNet apply(XLog eventLog, DiscoverBaseLineCNetPluginParametersImpl parameters) {
		HashMap<String, Double> activities = new HashMap<>();
		HashMap<Pair<String, String>, Double> dfg = new HashMap<>();
		HashMap<String, Integer> start = new HashMap<>();
		HashMap<String, Integer> end = new HashMap<>();
		XEventClassifier classifier = parameters.getEventClassifier();
		for (XTrace t : eventLog) {
			for (int i = 0; i < t.size(); i++) {
				XEvent event = t.get(i);
				String activity = classifier.getClassIdentity(event);
				if (!activities.containsKey(activity)) {
					activities.put(activity, 1d);
				} else {
					activities.put(activity, activities.get(activity) + 1);
				}
				if (i == 0) {
					if (!start.containsKey(activity)) {
						start.put(activity, 1);
					} else {
						start.put(activity, start.get(activity) + 1);
					}
				}
				if (i == t.size() - 1) {
					if (!end.containsKey(activity)) {
						end.put(activity, 1);
					} else {
						end.put(activity, end.get(activity) + 1);
					}
				}
				if (i > 0) {
					XEvent prev = t.get(i - 1);
					String prevAct = classifier.getClassIdentity(prev);
					Pair<String, String> df = new Pair<>(prevAct, activity);
					if (!dfg.containsKey(df)) {
						dfg.put(df, 1d);
					} else {
						dfg.put(df, dfg.get(df) + 1);
					}
				}
			}
		}
		CNetGenerator generator = new CNetGenerator(activities, dfg, start, end);
		return generator.generateModel(parameters.getDependencyThreshold(), parameters.getPositiveObservations(),
				parameters.getAndThreshold());
	}

}
